from abc import ABC, abstractmethod
from pathlib import Path
from typing import Callable, Dict, List, Union

from snowddl.config import SnowDDLConfig
from snowddl.blueprint import NameWithType
from snowddl.parser._parsed_file import ParsedFile


class AbstractParser(ABC):
    def __init__(self, config: SnowDDLConfig, base_path: Path):
        self.config = config
        self.base_path = base_path

        self.env_prefix = config.env_prefix

    @abstractmethod
    def load_blueprints(self):
        pass

    def parse_single_file(self, path: Path, json_schema: dict, callback: Callable[[ParsedFile],Union[None,Dict]] = None):
        if not callback:
            callback = lambda f: f.params

        if path.exists():
            try:
                file = ParsedFile(self, path, json_schema)
                return callback(file)
            except Exception as e:
                self.config.add_error(path, e)

        return {}

    def parse_schema_object_files(self, object_type: str, json_schema: dict, callback: Callable[[ParsedFile],None]):
        for path in self.base_path.glob(f"*/*/{object_type}/*.yaml"):
            try:
                file = ParsedFile(self, path, json_schema)
                callback(file)
            except Exception as e:
                self.config.add_error(path, e)

    def normalise_params_dict(self, params):
        if params is None:
            return None

        if isinstance(params, dict):
            return {k.upper(): v for k, v in params.items()}

        raise ValueError(f"Value is neither None, nor dictionary [{params}]")

    def validate_name_with_args(self, path: Path, arguments: List[NameWithType]):
        stem_name = str(path.stem)
        args_str = ','.join([a.type.base_type.name for a in arguments]).lower()

        open_pos = stem_name.find('(')
        close_pos = stem_name.find(')')

        if open_pos == -1 or close_pos == -1:
            raise ValueError(f"File [{path}] name should have list of arguments, e.g. [{stem_name}({args_str}).yaml]")

        base_name = stem_name[:open_pos]

        if stem_name[open_pos+1:close_pos] != args_str:
            raise ValueError(f"File [{path}] name does not match list of arguments, expected [{base_name}({args_str}).yaml]")

        return base_name



from snowddl.blueprint import NetworkPolicyBlueprint, Ident
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


network_policy_json_schema = {
    "type": "object",
    "additionalProperties": {
        "type": "object",
        "properties": {
            "allowed_ip_list": {
                "type": "array",
                "items": {
                    "type": "string"
                },
                "minItems": 1
            },
            "blocked_ip_list": {
                "type": "array",
                "items": {
                    "type": "string"
                },
                "minItems": 1
            },
            "comment": {
                "type": "string"
            }
        },
        "required": ["allowed_ip_list"],
        "additionalProperties": False
    }
}


class NetworkPolicyParser(AbstractParser):
    def load_blueprints(self):
        self.parse_single_file(self.base_path / 'network_policy.yaml', network_policy_json_schema, self.process_network_policy)

    def process_network_policy(self, file: ParsedFile):
        for policy_name, policy in file.params.items():
            bp = NetworkPolicyBlueprint(
                full_name=Ident(policy_name),
                allowed_ip_list=policy['allowed_ip_list'],
                blocked_ip_list=policy.get('blocked_ip_list', []),
                comment=policy.get('comment'),
            )

            self.config.add_blueprint(bp)



from re import compile, IGNORECASE

from snowddl.blueprint import ExternalTableBlueprint, ExternalTableColumn, PrimaryKeyBlueprint, UniqueKeyBlueprint, ForeignKeyBlueprint, DataType, Ident, SchemaObjectIdent, TableConstraintIdent, build_schema_object_ident
from snowddl.parser.abc_parser import AbstractParser, ParsedFile

col_type_re = compile(r'^(?P<type>[a-z0-9_]+(\((\d+)(,(\d+))?\))?)'
                      r'(?P<not_null> NOT NULL)?$'
                      , IGNORECASE)

external_table_json_schema = table_json_schema = {
    "type": "object",
    "properties": {
        "columns": {
            "type": "object",
            "additionalProperties": {
                "type": "object",
                "properties": {
                    "type": {
                        "type": "string"
                    },
                    "expr": {
                        "type": "string"
                    },
                    "comment": {
                        "type": "string"
                    }
                },
                "required": ["type", "expr"],
                "additionalProperties": False
            }
        },
        "location": {
            "type": "object",
            "properties": {
                "stage": {
                    "type": "string"
                },
                "path": {
                    "type": "string"
                },
                "pattern": {
                    "type": "string"
                },
                "file_format": {
                    "type": "string"
                }
            },
            "required": ["stage", "file_format"],
            "additionalProperties": False
        },
        "partition_by": {
            "type": "array",
            "items": {
                "type": "string"
            }
        },
        "partition_type": {
            "type": "string",
            "pattern": "^[A-Za-z_]+$"
        },
        "auto_refresh": {
            "type": "boolean"
        },
        "refresh_on_create": {
            "type": "boolean"
        },
        "aws_sns_topic": {
            "type": "string"
        },
        "table_format": {
            "type": "string",
            "pattern": "^[A-Za-z_]+$"
        },
        "integration": {
            "type": "string"
        },
        "comment": {
            "type": "string"
        },
        "primary_key": {
            "type": "array",
            "items": {
                "type": "string"
            },
            "minItems": 1
        },
        "unique_keys": {
            "type": "array",
            "items": {
                "type": "array",
                "items": {
                    "type": "string"
                },
                "minItems": 1,
            }
        },
        "foreign_keys": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "columns": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        },
                        "minItems": 1
                    },
                    "ref_table": {
                        "type": "string"
                    },
                    "ref_columns": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        },
                        "minItems": 1
                    }
                },
                "additionalProperties": False
            },
            "minItems": 1
        }
    },
    "required": ["location"],
    "additionalProperties": False
}


class ExternalTableParser(AbstractParser):
    def load_blueprints(self):
        self.parse_schema_object_files('external_table', external_table_json_schema, self.process_external_table)

    def process_external_table(self, f: ParsedFile):
        column_blueprints = []

        for col_name, col in f.params['columns'].items():
            m = col_type_re.match(col['type'])

            if not m:
                raise ValueError(f"Incorrect short syntax for column [{col_name}] in table [{f.database}.{f.schema}.{f.name}]: {col['type']}")

            column_blueprints.append(
                ExternalTableColumn(
                    name=Ident(col_name),
                    type=DataType(m.group('type')),
                    expr=col['expr'],
                    not_null=bool(m.group('not_null')),
                    comment=col.get('comment'),
                )
            )

        bp = ExternalTableBlueprint(
            full_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
            columns=column_blueprints if column_blueprints else None,
            partition_by=[Ident(col_name) for col_name in f.params['partition_by']] if f.params.get('partition_by') else None,
            partition_type=f.params['partition_type'].upper() if f.params.get('partition_type') else None,
            location_stage=build_schema_object_ident(self.env_prefix, f.params['location']['stage'], f.database, f.schema),
            location_path=f.params['location'].get('path'),
            location_pattern=f.params['location'].get('pattern'),
            file_format=build_schema_object_ident(self.env_prefix, f.params['location'].get('file_format'), f.database, f.schema) if f.params['location'].get('file_format') else None,
            refresh_on_create=f.params.get('refresh_on_create', False),
            auto_refresh=f.params.get('auto_refresh', False),
            aws_sns_topic=f.params.get('aws_sns_topic'),
            table_format=f.params['table_format'].upper() if f.params.get('table_format') else None,
            integration=Ident(f.params['integration']) if f.params.get('integration') else None,
            comment=f.params.get('comment'),
        )

        self.config.add_blueprint(bp)

        # Constraints

        if f.params.get('primary_key'):
            bp = PrimaryKeyBlueprint(
                full_name=TableConstraintIdent(self.env_prefix, f.database, f.schema, f.name, columns=[Ident(c) for c in f.params['primary_key']]),
                table_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
                columns=[Ident(c) for c in f.params['primary_key']],
                comment=None,
            )

            self.config.add_blueprint(bp)

        for columns in f.params.get('unique_keys', []):
            bp = UniqueKeyBlueprint(
                full_name=TableConstraintIdent(self.env_prefix, f.database, f.schema, f.name, columns=[Ident(c) for c in columns]),
                table_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
                columns=[Ident(c) for c in columns],
                comment=None,
            )

            self.config.add_blueprint(bp)

        for fk in f.params.get('foreign_keys', []):
            bp = ForeignKeyBlueprint(
                full_name=TableConstraintIdent(self.env_prefix, f.database, f.schema, f.name, columns=[Ident(c) for c in fk['columns']]),
                table_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
                columns=[Ident(c) for c in fk['columns']],
                ref_table_name=build_schema_object_ident(self.env_prefix, fk['ref_table'], f.database, f.schema),
                ref_columns=[Ident(c) for c in fk['ref_columns']],
                comment=None,
            )

            self.config.add_blueprint(bp)



from snowddl.blueprint import RoleBlueprint, DatabaseShareBlueprint, Grant, build_role_ident, build_grant_name_ident_snowflake
from snowddl.resolver.abc_role_resolver import AbstractRoleResolver, ObjectType


class InboundShareRoleResolver(AbstractRoleResolver):
    skip_on_empty_blueprints = True

    def get_role_suffix(self):
        return self.config.INBOUND_SHARE_ROLE_SUFFIX

    def get_existing_role_grants(self, role_name):
        # There is no way to detect grant of IMPORTED PRIVILEGES directly
        # SHOW GRANTS output is filled with whatever grants come with INBOUND SHARE
        # IMPORTED PRIVILEGES grant has to be emulated
        grants = []

        cur = self.engine.execute_meta("SHOW GRANTS TO ROLE {role_name:i}", {
            "role_name": role_name,
        })

        for r in cur:
            # Convert DATABASE.USAGE grant into IMPORTED PRIVILEGES
            # Ignore everything else
            if r['privilege'] == 'USAGE' and r['granted_on'] == 'DATABASE':
                grants.append(Grant(
                    privilege="IMPORTED PRIVILEGES",
                    on=ObjectType.DATABASE,
                    name=build_grant_name_ident_snowflake(r['name'], ObjectType.DATABASE),
                ))

                break

        return role_name, grants, []

    def get_blueprints(self):
        blueprints = []

        for user in self.config.get_blueprints_by_type(DatabaseShareBlueprint).values():
            blueprints.append(self.get_blueprint_inbound_share_role(user))

        return {str(bp.full_name): bp for bp in blueprints}

    def get_blueprint_inbound_share_role(self, bp: DatabaseShareBlueprint):
        grants = []

        grants.append(Grant(
            privilege="IMPORTED PRIVILEGES",
            on=ObjectType.DATABASE,
            name=bp.full_name,
        ))

        return RoleBlueprint(
            full_name=build_role_ident(self.config.env_prefix, bp.full_name.database, self.get_role_suffix()),
            grants=grants,
            future_grants=[],
            comment=None,
        )



from dataclasses import dataclass, field
from typing import List

from snowddl.blueprint import DatabaseIdent, ObjectType

@dataclass
class SnowDDLSettings:
    execute_safe_ddl: bool = False
    execute_unsafe_ddl: bool = False
    execute_replace_table: bool = False
    execute_masking_policy: bool = False
    execute_row_access_policy: bool = False
    execute_account_params: bool = False
    execute_network_policy: bool = False
    execute_resource_monitor: bool = False
    execute_inbound_share: bool = False
    execute_outbound_share: bool = False
    refresh_user_passwords: bool = False
    refresh_future_grants: bool = False
    exclude_object_types: List[ObjectType] = field(default_factory=list)
    include_object_types: List[ObjectType] = field(default_factory=list)
    include_databases: List[DatabaseIdent] = field(default_factory=list)
    ignore_ownership: bool = False
    max_workers: int = 8



from snowddl.blueprint import StageBlueprint
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType

class StageResolver(AbstractSchemaObjectResolver):
    def get_object_type(self) -> ObjectType:
        return ObjectType.STAGE

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta("SHOW STAGES IN SCHEMA {database:i}.{schema:i}", {
            "database": schema['database'],
            "schema": schema['schema'],
        })

        for r in cur:
            if 'TEMPORARY' in r['type']:
                # Skip TEMPORARY stages
                continue

            existing_objects[f"{r['database_name']}.{r['schema_name']}.{r['name']}"] = {
                "database": r['database_name'],
                "schema": r['schema_name'],
                "name": r['name'],
                "url": r['url'],
                "storage_integration": r['storage_integration'],
                "owner": r['owner'],
                "type": r['type'],
                "comment": r['comment'],
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(StageBlueprint)

    def create_object(self, bp: StageBlueprint):
        common_query = self._build_common_stage_sql(bp)
        create_query = self.engine.query_builder()

        create_query.append("CREATE STAGE {full_name:i}", {
            "full_name": bp.full_name,
        })

        create_query.append(common_query)

        self.engine.execute_safe_ddl(create_query)

        self.engine.execute_safe_ddl("COMMENT ON STAGE {full_name:i} IS {comment}", {
            "full_name": bp.full_name,
            "comment": common_query.add_short_hash(bp.comment),
        })

        return ResolveResult.CREATE

    def compare_object(self, bp: StageBlueprint, row: dict):
        common_query = self._build_common_stage_sql(bp)

        if not common_query.compare_short_hash(row['comment']):
            alter_query = self.engine.query_builder()

            alter_query.append("ALTER STAGE {full_name:i} SET", {
                "full_name": bp.full_name,
            })

            alter_query.append(common_query)

            self.engine.execute_safe_ddl(alter_query)

            self.engine.execute_safe_ddl("COMMENT ON STAGE {full_name:i} IS {comment}", {
                "full_name": bp.full_name,
                "comment": common_query.add_short_hash(bp.comment),
            })

            return ResolveResult.ALTER

        return ResolveResult.NOCHANGE

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl("DROP STAGE {database:i}.{schema:i}.{name:i}", {
            "database": row['database'],
            "schema": row['schema'],
            "name": row['name'],
        })

        return ResolveResult.DROP

    def _build_common_stage_sql(self, bp: StageBlueprint):
        query = self.engine.query_builder()

        if bp.url:
            query.append_nl("URL = {url}", {
                "url": bp.url
            })

        if bp.storage_integration:
            query.append_nl("STORAGE_INTEGRATION = {storage_integration:i}", {
                "storage_integration": bp.storage_integration,
            })

        if bp.encryption:
            query.append_nl("ENCRYPTION = (")

            for k, v in bp.encryption.items():
                query.append("{param_name:r} = {param_value:dp}", {
                    "param_name": k,
                    "param_value": v,
                })

            query.append(")")

        if bp.directory:
            query.append_nl("DIRECTORY = (")

            for k, v in bp.directory.items():
                query.append("{param_name:r} = {param_value:dp}", {
                    "param_name": k,
                    "param_value": v,
                })

            query.append(")")

        if bp.file_format:
            query.append_nl("FILE_FORMAT = (FORMAT_NAME = {file_format:i})", {
                "file_format": bp.file_format,
            })

        if bp.copy_options:
            query.append_nl("COPY_OPTIONS = (")

            for k, v in bp.copy_options.items():
                query.append("{option_name:r} = {option_value:dp}", {
                    "option_name": k,
                    "option_value": v,
                })

            query.append(")")

        if bp.comment:
            query.append_nl("COMMENT = {comment}", {
                "comment": bp.comment,
            })

        return query



from snowddl.blueprint import DatabaseShareBlueprint, InboundShareIdent
from snowddl.resolver.abc_resolver import AbstractResolver, ResolveResult, ObjectType, SnowDDLUnsupportedError


class InboundShareResolver(AbstractResolver):
    skip_on_empty_blueprints = True

    def get_object_type(self) -> ObjectType:
        return ObjectType.DATABASE

    def get_existing_objects(self):
        existing_objects = {}

        # Inbound shares do not support env_prefix due to limitations of Snowflake
        # SHOW DATABASES without filter is not efficient, but still better than alternatives
        # Hopefully, Snowflake will provide a dedicated SHOW command for DATABASES created from INBOUND SHARES in future
        cur = self.engine.execute_meta("SHOW DATABASES")

        for r in cur:
            # Skip databases created by other roles
            if r['owner'] != self.engine.context.current_role:
                continue

            # Skip normal databases (not created from SHARES)
            if not r['origin']:
                continue

            existing_objects[r['name']] = {
                "database": r['name'],
                "share": r['origin'],
                "comment": r['comment'] if r['comment'] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(DatabaseShareBlueprint)

    def create_object(self, bp: DatabaseShareBlueprint):
        query = self.engine.query_builder()

        query.append("CREATE DATABASE {full_name:i} FROM SHARE {share_name:i}", {
            "full_name": bp.full_name,
            "share_name": bp.share_name,
        })

        if bp.comment:
            query.append_nl("COMMENT = {comment}", {
                "comment": bp.comment,
            })

        self.engine.execute_safe_ddl(query, condition=self.engine.settings.execute_inbound_share)

        return ResolveResult.CREATE

    def compare_object(self, bp: DatabaseShareBlueprint, row: dict):
        query = self.engine.query_builder()

        # Database is attached to another share, re-create it
        if str(bp.share_name) != row['share']:
            query.append("CREATE OR REPLACE DATABASE {full_name:i} FROM SHARE {share_name:i}", {
                "full_name": bp.full_name,
                "share_name": bp.share_name,
            })

            if bp.comment:
                query.append_nl("COMMENT = {comment}", {
                    "comment": bp.comment,
                })

            self.engine.execute_unsafe_ddl(query, condition=self.engine.settings.execute_inbound_share)
            return ResolveResult.REPLACE

        if bp.comment != row['comment']:
            self.engine.execute_safe_ddl("ALTER DATABASE {full_name:i} SET COMMENT = {comment}", {
                "full_name": bp.full_name,
                "comment": bp.comment,
            }, condition=self.engine.settings.execute_inbound_share)

            return ResolveResult.ALTER

        return ResolveResult.NOCHANGE

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl("DROP DATABASE {database:i}", {
            "database": row['database']
        }, condition=self.engine.settings.execute_inbound_share)

        return ResolveResult.DROP



from snowddl.blueprint import ForeignKeyBlueprint
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType


class ForeignKeyResolver(AbstractSchemaObjectResolver):
    def get_object_type(self) -> ObjectType:
        return ObjectType.FOREIGN_KEY

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}
        constraints_by_name = {}

        cur = self.engine.execute_meta("SHOW IMPORTED KEYS IN SCHEMA {database:i}.{schema:i}", {
            "database": schema['database'],
            "schema": schema['schema'],
        })

        for r in cur:
            if r['fk_name'] not in constraints_by_name:
                constraints_by_name[r['fk_name']] = {
                    "database": r['fk_database_name'],
                    "schema": r['fk_schema_name'],
                    "table": r['fk_table_name'],
                    "columns": {r['key_sequence']: r['fk_column_name']},
                    "ref_database": r['pk_database_name'],
                    "ref_schema": r['pk_schema_name'],
                    "ref_table": r['pk_table_name'],
                    "ref_columns": {r['key_sequence']: r['pk_column_name']},
                }
            else:
                constraints_by_name[r['fk_name']]['columns'][r['key_sequence']] = r['fk_column_name']
                constraints_by_name[r['fk_name']]['ref_columns'][r['key_sequence']] = r['pk_column_name']

        for c in constraints_by_name.values():
            columns_list = [c['columns'][k] for k in sorted(c['columns'])]
            ref_columns_list = [c['ref_columns'][k] for k in sorted(c['ref_columns'])]

            full_name = f"{c['database']}.{c['schema']}.{c['table']}({','.join(columns_list)})"

            existing_objects[full_name] = {
                "database": c['database'],
                "schema": c['schema'],
                "table": c['table'],
                "columns": columns_list,
                "ref_table_name": f"{c['ref_database']}.{c['ref_schema']}.{c['ref_table']}",
                "ref_columns": ref_columns_list,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(ForeignKeyBlueprint)

    def create_object(self, bp: ForeignKeyBlueprint):
        self.engine.execute_safe_ddl("ALTER TABLE {table_name:i} ADD FOREIGN KEY ({columns:i}) REFERENCES {ref_table_name:i} ({ref_columns:i})", {
            "table_name": bp.table_name,
            "columns": bp.columns,
            "ref_table_name": bp.ref_table_name,
            "ref_columns": bp.ref_columns,
        })

        return ResolveResult.CREATE

    def compare_object(self, bp: ForeignKeyBlueprint, row: dict):
        if str(bp.ref_table_name) == row['ref_table_name'] and [str(c) for c in bp.ref_columns] == row['ref_columns']:
            return ResolveResult.NOCHANGE

        self.engine.execute_safe_ddl("ALTER TABLE {table_name:i} DROP FOREIGN KEY ({columns:i})", {
            "table_name": bp.table_name,
            "columns": bp.columns,
        })

        self.engine.execute_safe_ddl("ALTER TABLE {table_name:i} ADD FOREIGN KEY ({columns:i}) REFERENCES {ref_table_name:i} ({ref_columns:i})", {
            "table_name": bp.table_name,
            "columns": bp.columns,
            "ref_table_name": bp.ref_table_name,
            "ref_columns": bp.ref_columns,
        })

        return ResolveResult.ALTER

    def drop_object(self, row: dict):
        self.engine.execute_safe_ddl("ALTER TABLE {database:i}.{schema:i}.{table:i} DROP FOREIGN KEY ({columns:i})", {
            "database": row['database'],
            "schema": row['schema'],
            "table": row['table'],
            "columns": row['columns'],
        })

        return ResolveResult.DROP



from argparse import ArgumentParser, HelpFormatter
from dataclasses import fields, is_dataclass
from os import environ, getcwd
from typing import Optional

from snowddl.app.base import BaseApp
from snowddl.blueprint import AbstractBlueprint, DatabaseBlueprint, DatabaseIdent, SchemaIdent, SchemaObjectIdent, SchemaBlueprint, SchemaObjectBlueprint
from snowddl.config import SnowDDLConfig
from snowddl.parser import singledb_parser_sequence
from snowddl.resolver import singledb_resolver_sequence


class SingleDbApp(BaseApp):
    parser_sequence = singledb_parser_sequence
    resolver_sequence = singledb_resolver_sequence

    def __init__(self):
        self.config_db: Optional[DatabaseIdent] = None
        self.target_db: Optional[DatabaseIdent] = None

        super().__init__()

    def init_arguments_parser(self):
        formatter = lambda prog: HelpFormatter(prog, max_help_position=32)
        parser = ArgumentParser(prog='snowddl-singledb', description='Special SnowDDL mode to process schema objects of single database only', formatter_class=formatter)

        # Config
        parser.add_argument('-c', help='Path to config directory OR name of bundled test config (default: current directory)', metavar='CONFIG_PATH', default=getcwd())

        # Auth
        parser.add_argument('-a', help='Snowflake account identifier (default: SNOWFLAKE_ACCOUNT env variable)', metavar='ACCOUNT', default=environ.get('SNOWFLAKE_ACCOUNT'))
        parser.add_argument('-u', help='Snowflake user name (default: SNOWFLAKE_USER env variable)', metavar='USER', default=environ.get('SNOWFLAKE_USER'))
        parser.add_argument('-p', help='Snowflake user password (default: SNOWFLAKE_PASSWORD env variable)', metavar='PASSWORD', default=environ.get('SNOWFLAKE_PASSWORD'))
        parser.add_argument('-k', help='Path to private key file (default: SNOWFLAKE_PRIVATE_KEY_PATH env variable)', metavar='PRIVATE_KEY', default=environ.get('SNOWFLAKE_PRIVATE_KEY_PATH'))

        # Role & warehouse
        parser.add_argument('-r', help='Snowflake active role (default: SNOWFLAKE_ROLE env variable)', metavar='ROLE', default=environ.get('SNOWFlAKE_ROLE'))
        parser.add_argument('-w', help='Snowflake active warehouse (default: SNOWFLAKE_WAREHOUSE env variable)', metavar='WAREHOUSE', default=environ.get('SNOWFLAKE_WAREHOUSE'))

        # SingleDb options
        parser.add_argument('--config-db', help='Source database name in config (default: detected automatically if only one database is present in config)', default=None)
        parser.add_argument('--target-db', help='Target database name in Snowflake account (default: same as --config-db)', default=None)

        # Generic options
        parser.add_argument('--passphrase', help='Passphrase for private key file (default: SNOWFLAKE_PRIVATE_KEY_PASSPHRASE env variable)', default=environ.get('SNOWFLAKE_PRIVATE_KEY_PASSPHRASE'))
        parser.add_argument('--env-prefix', help='Env prefix added to global object names, used to separate environments (e.g. DEV, PROD)', default=environ.get('SNOWFLAKE_ENV_PREFIX'))
        parser.add_argument('--max-workers', help='Maximum number of workers to resolve objects in parallel', default=None, type=int)

        # Logging
        parser.add_argument('--log-level', help="Log level (possible values: DEBUG, INFO, WARNING; default: INFO)", default="INFO")
        parser.add_argument('--show-sql', help="Show executed DDL queries", default=False, action='store_true')

        # Placeholders
        parser.add_argument('--placeholder-path', help='Path to config file with environment-specific placeholders', default=None, metavar='')
        parser.add_argument('--placeholder-values', help='Environment-specific placeholder values in JSON format', default=None, metavar='')

        # Object types
        parser.add_argument('--exclude-object-types', help="Comma-separated list of object types NOT to resolve", default=None, metavar='')
        parser.add_argument('--include-object-types', help="Comma-separated list of object types TO resolve, all other types are excluded", default=None, metavar='')

        # Apply even more unsafe changes
        parser.add_argument('--apply-unsafe', help="Additionally apply unsafe changes, which may cause loss of data (ALTER, DROP, etc.)", default=False, action='store_true')
        parser.add_argument('--apply-replace-table', help="Additionally apply REPLACE TABLE when ALTER TABLE is not possible", default=False, action='store_true')
        parser.add_argument('--apply-masking-policy', help="Additionally apply changes to MASKING POLICIES", default=False, action='store_true')
        parser.add_argument('--apply-row-access-policy', help="Additionally apply changes to ROW ACCESS POLICIES", default=False, action='store_true')

        # Subparsers
        subparsers = parser.add_subparsers(dest="action")
        subparsers.required = True

        subparsers.add_parser('plan', help="Resolve objects, apply nothing, display suggested changes")
        subparsers.add_parser('apply', help="Resolve objects, apply safe changes, display suggested unsafe changes")
        subparsers.add_parser('destroy', help="Drop objects with specified --env-prefix, use it to reset dev and test environments")

        return parser

    def init_config(self):
        config = super().init_config()

        # Init Source DB
        if self.args.get('config_db'):
            self.config_db = DatabaseIdent(config.env_prefix, self.args.get('config_db'))

            if str(self.config_db) not in config.get_blueprints_by_type(DatabaseBlueprint):
                raise ValueError(f"Source database [{self.config_db}] does not exist in config")
        else:
            if len(config.get_blueprints_by_type(DatabaseBlueprint)) > 1:
                raise ValueError(f"More than one source database exist in config, please choose a specific database using --source-db argument")

            self.config_db = next(iter(config.get_blueprints_by_type(DatabaseBlueprint).values())).full_name

        # Init Target DB
        if self.args.get('target_db'):
            self.target_db = DatabaseIdent(config.env_prefix, self.args.get('target_db'))
        else:
            self.target_db = self.config_db

        return self.convert_config(config)

    def convert_config(self, original_config: SnowDDLConfig):
        singledb_config = SnowDDLConfig(self.args.get('env_prefix'))

        for bp_dict in original_config.blueprints.values():
            for bp in bp_dict.values():
                if isinstance(bp, (SchemaBlueprint, SchemaObjectBlueprint)) and bp.full_name.database_full_name == self.config_db:
                    singledb_config.add_blueprint(self.convert_blueprint(bp))

        return singledb_config

    def convert_blueprint(self, bp: AbstractBlueprint):
        for f in fields(bp):
            self.convert_object_recursive(getattr(bp, f.name))

        return bp

    def convert_object_recursive(self, obj):
        if is_dataclass(obj):
            for f in fields(obj):
                self.convert_object_recursive(getattr(obj, f.name))

        if isinstance(obj, list):
            for item in obj:
                self.convert_object_recursive(item)

        if isinstance(obj, dict):
            for item in obj.values():
                self.convert_object_recursive(item)

        if isinstance(obj, (SchemaIdent, SchemaObjectIdent)):
            obj.database = self.target_db.database

        return obj

    def init_settings(self):
        settings = super().init_settings()
        settings.include_databases = [self.target_db]
        settings.ignore_ownership = True

        return settings


def entry_point():
    app = SingleDbApp()
    app.execute()



from snowddl.blueprint import AccountIdent, OutboundShareBlueprint, OutboundShareIdent, Grant, ObjectType
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


outbound_share_json_schema = {
    "type": "object",
    "additionalProperties": {
        "properties": {
            "accounts": {
                "type": "array",
                "items": {
                    "type": "string"
                },
            },
            "share_restrictions": {
                "type": "boolean"
            },
            "grants": {
                "type": "object",
                "additionalProperties": {
                    "type": "array",
                    "items": {
                        "type": "string"
                    },
                    "minItems": 1
                }
            },
            "comment": {
                "type": "string"
            }
        },
        "required": ["grants"],
        "additionalProperties": False
    }
}


class OutboundShareParser(AbstractParser):
    def load_blueprints(self):
        self.parse_single_file(self.base_path / 'outbound_share.yaml', outbound_share_json_schema, self.process_inbound_share)

    def process_inbound_share(self, f: ParsedFile):
        for share_name, share in f.params.items():
            bp = OutboundShareBlueprint(
                full_name=OutboundShareIdent(self.env_prefix, share_name),
                accounts=self.get_share_accounts(share),
                share_restrictions=share.get('share_restrictions'),
                grants=self.get_share_grants(share),
                comment=share.get('comment'),
            )

            self.config.add_blueprint(bp)

    def get_share_accounts(self, share):
        share_accounts = []

        for account in share.get('accounts', []):
            account_parts = account.split('.')

            if len(account_parts) != 2:
                raise ValueError(f"Invalid outbound share account format [{account}], expected identifier with exactly 2 parts: <organization>.<account>")

            share_accounts.append(AccountIdent(*account_parts))

        return share_accounts

    def get_share_grants(self, share):
        grants = []

        for definition, pattern_list in share['grants'].items():
            on, privileges = definition.upper().split(':')

            for p in privileges.split(','):
                for pattern in pattern_list:
                    blueprints = self.config.get_blueprints_by_type_and_pattern(ObjectType[on].blueprint_cls, pattern)

                    if not blueprints:
                        raise ValueError(f"No {ObjectType[on].plural} matched wildcard grant with pattern [{pattern}]")

                    for object_bp in blueprints.values():
                        grants.append(Grant(
                            privilege=p,
                            on=ObjectType[on],
                            name=object_bp.full_name,
                        ))

        return grants



from snowddl.blueprint import PrimaryKeyBlueprint
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType


class PrimaryKeyResolver(AbstractSchemaObjectResolver):
    def get_object_type(self) -> ObjectType:
        return ObjectType.PRIMARY_KEY

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}
        constraints_by_name = {}

        cur = self.engine.execute_meta("SHOW PRIMARY KEYS IN SCHEMA {database:i}.{schema:i}", {
            "database": schema['database'],
            "schema": schema['schema'],
        })

        for r in cur:
            if r['constraint_name'] not in constraints_by_name:
                constraints_by_name[r['constraint_name']] = {
                    "database": r['database_name'],
                    "schema": r['schema_name'],
                    "table": r['table_name'],
                    "columns": {r['key_sequence']: r['column_name']}
                }
            else:
                constraints_by_name[r['constraint_name']]['columns'][r['key_sequence']] = r['column_name']

        for c in constraints_by_name.values():
            columns_list = [c['columns'][k] for k in sorted(c['columns'])]
            full_name = f"{c['database']}.{c['schema']}.{c['table']}({','.join(columns_list)})"

            existing_objects[full_name] = {
                "database": c['database'],
                "schema": c['schema'],
                "table": c['table'],
                "columns": columns_list,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(PrimaryKeyBlueprint)

    def create_object(self, bp: PrimaryKeyBlueprint):
        self.engine.execute_safe_ddl("ALTER TABLE {table_name:i} ADD PRIMARY KEY ({columns:i})", {
            "table_name": bp.table_name,
            "columns": bp.columns,
        })

        return ResolveResult.CREATE

    def compare_object(self, bp: PrimaryKeyBlueprint, row: dict):
        if [str(c) for c in bp.columns] == row['columns']:
            return ResolveResult.NOCHANGE

        self.engine.execute_safe_ddl("ALTER TABLE {table_name:i} DROP PRIMARY KEY", {
            "table_name": bp.table_name,
        })

        self.engine.execute_safe_ddl("ALTER TABLE {table_name:i} ADD PRIMARY KEY ({columns:i})", {
            "table_name": bp.table_name,
            "columns": bp.columns,
        })

        return ResolveResult.ALTER

    def drop_object(self, row: dict):
        self.engine.execute_safe_ddl("ALTER TABLE {database:i}.{schema:i}.{table:i} DROP PRIMARY KEY", {
            "database": row['database'],
            "schema": row['schema'],
            "table": row['table'],
        })

        return ResolveResult.DROP



from snowddl.blueprint import Grant, BusinessRoleBlueprint, AbstractIdentWithPrefix, Ident, ObjectType, SchemaBlueprint, build_role_ident
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


business_role_json_schema = {
    "type": "object",
    "additionalProperties": {
        "type": "object",
        "properties": {
            "schema_owner": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            },
            "schema_read": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            },
            "schema_write": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            },
            "warehouse_usage": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            },
            "warehouse_monitor": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            },
            "tech_roles": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            },
            "global_roles": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            },
            "comment": {
                "type": "string"
            }
        }
    }
}


class BusinessRoleParser(AbstractParser):
    def load_blueprints(self):
        self.parse_single_file(self.base_path / 'business_role.yaml', business_role_json_schema, self.process_business_role)

    def process_business_role(self, file: ParsedFile):
        for business_role_name, business_role in file.params.items():
            business_role_ident = build_role_ident(self.env_prefix, business_role_name, self.config.BUSINESS_ROLE_SUFFIX)

            grants = []

            for full_schema_name in business_role.get('schema_owner', []):
                grants.extend(self.build_schema_role_grants(full_schema_name, 'OWNER'))

            for full_schema_name in business_role.get('schema_read', []):
                grants.extend(self.build_schema_role_grants(full_schema_name, 'READ'))

            for full_schema_name in business_role.get('schema_write', []):
                grants.extend(self.build_schema_role_grants(full_schema_name, 'WRITE'))

            for warehouse_name in business_role.get('warehouse_usage', []):
                grants.append(self.build_warehouse_role_grant(warehouse_name, 'USAGE'))

            for warehouse_name in business_role.get('warehouse_monitor', []):
                grants.append(self.build_warehouse_role_grant(warehouse_name, 'MONITOR'))

            for tech_role_name in business_role.get('tech_roles', []):
                grants.append(Grant(
                    privilege="USAGE",
                    on=ObjectType.ROLE,
                    name=build_role_ident(self.env_prefix, tech_role_name, self.config.TECH_ROLE_SUFFIX),
                ))

            for global_role_name in business_role.get('global_roles', []):
                grants.append(Grant(
                    privilege="USAGE",
                    on=ObjectType.ROLE,
                    name=Ident(global_role_name),
                ))

            bp = BusinessRoleBlueprint(
                full_name=business_role_ident,
                grants=grants,
                future_grants=[],
                comment=business_role.get('comment'),
            )

            self.config.add_blueprint(bp)

    def build_schema_role_grants(self, full_schema_name, grant_type):
        grants = []

        for schema_bp in self.config.get_blueprints_by_type_and_pattern(SchemaBlueprint, full_schema_name).values():
            grants.append(
                Grant(
                    privilege="USAGE",
                    on=ObjectType.ROLE,
                    name=build_role_ident(self.env_prefix, schema_bp.full_name.database, schema_bp.full_name.schema, grant_type, self.config.SCHEMA_ROLE_SUFFIX),
                )
            )

        if not grants:
            raise ValueError(f"No {ObjectType.SCHEMA.plural} matched wildcard grant with pattern [{full_schema_name}]")

        return grants

    def build_warehouse_role_grant(self, warehouse_name, grant_type):
        return Grant(
            privilege="USAGE",
            on=ObjectType.ROLE,
            name=build_role_ident(self.env_prefix, warehouse_name, grant_type, self.config.WAREHOUSE_ROLE_SUFFIX),
        )



import re
import string

from snowddl.blueprint import AbstractIdent


class SnowDDLFormatter(string.Formatter):
    safe_ident_regexp = re.compile(r'^[A-Za-z_][A-Za-z0-9_]*$')
    safe_decimal_regexp = re.compile(r'^[+-]?[0-9]+(\.[0-9]+)?$')
    safe_float_regexp = re.compile(r'^[+-]?[0-9]+(\.[0-9]+([e|E][+-][0-9]+)?)?$')

    def __init__(self):
        self.smart_transformations = {
            's': self.quote,
            'd': self.safe_decimal,
            'f': self.safe_float,
            'b': self.safe_bool,
            'i': self.quote_ident,
            'in': self.quote_ident_no_argument,
            'r': str,
            'lf': self.quote_like_full_match,
            'ls': self.quote_like_starts_with,
            'le': self.quote_like_ends_with,
            'lse': self.quote_like_starts_ends_with,
        }

        self.raw_transformations = {
            'dp': self.dynamic_param,
        }

        self.default_transformation = 's'

    def format_sql(self, sql, params=None):
        sql = str(sql)

        if params:
            return self.vformat(sql, [], params)

        return sql

    def convert_field(self, value, conversion):
        if conversion is not None:
            raise ValueError("Conversions are disabled for SnowDDLFormatter")

        return value

    def format_field(self, value, format_spec):
        if not format_spec:
            format_spec = self.default_transformation

        if format_spec in self.raw_transformations:
            return self.raw_transformations[format_spec](value)

        if format_spec not in self.smart_transformations:
            raise ValueError(f"Unknown format transformation [{format_spec}]")

        if isinstance(value, list) and format_spec != 'dp':
            if not value:
                raise ValueError("Attempt to format an empty list")
            return ', '.join([self.smart_transformations[format_spec](v) for v in value])
        else:
            return self.smart_transformations[format_spec](value)

    @classmethod
    def escape(cls, val):
        return str(val).replace("'", "''").replace("\\", "\\\\")

    @classmethod
    def escape_ident(cls, val):
        val = str(val)

        if not val:
            raise ValueError("Identifier cannot be empty")

        return val.replace('"', '""')

    @classmethod
    def escape_like(cls, val):
        return str(val).replace("_", "\\_").replace("%", "\\%")

    @classmethod
    def quote(cls, val):
        if val is None:
            return 'NULL'

        return f"'{cls.escape(val)}'"

    @classmethod
    def quote_ident(cls, val):
        if not isinstance(val, AbstractIdent):
            return f'"{cls.escape_ident(val)}"'

        core_parts, argument_parts = val.parts_for_format()

        if argument_parts is not None:
            return '.'.join(f'"{cls.escape_ident(p)}"' for p in core_parts) + \
                   '(' + ','.join(cls.safe_ident(p) for p in argument_parts) + ')'

        return '.'.join(f'"{cls.escape_ident(p)}"' for p in core_parts)

    @classmethod
    def quote_ident_no_argument(cls, val):
        if not isinstance(val, AbstractIdent):
            return f'"{cls.escape_ident(val)}"'

        core_parts, _ = val.parts_for_format()

        return '.'.join(f'"{cls.escape_ident(p)}"' for p in core_parts)

    @classmethod
    def quote_like_full_match(cls, val):
        return cls.quote(f"{cls.escape_like(val)}")

    @classmethod
    def quote_like_starts_with(cls, val):
        return cls.quote(f"{cls.escape_like(val)}%")

    @classmethod
    def quote_like_ends_with(cls, val):
        return cls.quote(f"%{cls.escape_like(val)}")

    @classmethod
    def quote_like_starts_ends_with(cls, val):
        return cls.quote(f"{cls.escape_like(val[0])}%{cls.escape_like(val[1])}")

    @classmethod
    def safe_ident(cls, val):
        val = str(val)

        if not cls.safe_ident_regexp.match(val):
            raise ValueError(f"Value [{val}] is not a safe identifier")

        return val.upper()

    @classmethod
    def safe_float(cls, val):
        if val is None:
            return 'NULL'

        val = str(val)

        if not cls.safe_float_regexp.match(val):
            raise ValueError(f'Value [{val}] is not a safe float')

        return val

    @classmethod
    def safe_decimal(cls, val):
        if val is None:
            return 'NULL'

        val = str(val)

        if not cls.safe_decimal_regexp.match(val):
            raise ValueError(f'Value [{val}] is not a safe integer')

        return val

    @classmethod
    def safe_bool(cls, val):
        if not isinstance(val, bool):
            raise ValueError(f'Value [{val}] is not a safe boolean')

        return 'TRUE' if val else 'FALSE'

    @classmethod
    def dynamic_param(cls, val):
        if isinstance(val, list):
            return f"({', '.join([cls.dynamic_param(v) for v in val])})"
        elif isinstance(val, bool):
            return cls.safe_bool(val)
        elif isinstance(val, int):
            return cls.safe_decimal(val)
        elif isinstance(val, float):
            return cls.safe_float(val)
        else:
            return cls.quote(val)



from snowddl.blueprint import MaterializedViewBlueprint, ViewColumn, Ident, SchemaObjectIdent
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


materialized_view_json_schema = {
    "type": "object",
    "properties": {
        "columns": {
            "type": "object",
            "additionalProperties": {
                "type": "string"
            }
        },
        "text": {
            "type": "string"
        },
        "cluster_by": {
            "type": "array",
            "items": {
                "type": "string"
            },
            "minItems": 1
        },
        "is_secure": {
            "type": "boolean"
        },
        "comment": {
            "type": "string"
        }
    },
    "required": ["text"],
    "additionalProperties": False
}


class MaterializedViewParser(AbstractParser):
    def load_blueprints(self):
        self.parse_schema_object_files("materialized_view", materialized_view_json_schema, self.process_materialized_view)

    def process_materialized_view(self, f: ParsedFile):
        column_blueprints = []

        for col_name, col_comment in f.params.get('columns', {}).items():
            column_blueprints.append(
                ViewColumn(
                    name=Ident(col_name),
                    comment=col_comment if col_comment else None,
                )
            )

        bp = MaterializedViewBlueprint(
            full_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
            text=f.params['text'],
            columns=column_blueprints if column_blueprints else None,
            is_secure=f.params.get('is_secure', False),
            cluster_by=f.params.get('cluster_by'),
            comment=f.params.get('comment'),
        )

        self.config.add_blueprint(bp)
